#!/usr/bin/env python

# Edit this script to add your team's training code.
# Some functions are *required*, but you can edit most parts of the required functions, remove non-required functions, and add your own functions.

################################################################################
#
# Imported functions and variables
#
################################################################################

# Import functions. These functions are not required. You can change or remove them.
from helper_code import *
import numpy as np, os, sys, joblib
import pandas as pd
from sklearn.impute import SimpleImputer
from sklearn.ensemble import RandomForestClassifier
import time

import utils


# Define the Challenge lead sets. These variables are not required. You can change or remove them.
twelve_leads = ('I', 'II', 'III', 'aVR', 'aVL', 'aVF', 'V1', 'V2', 'V3', 'V4', 'V5', 'V6')
six_leads = ('I', 'II', 'III', 'aVR', 'aVL', 'aVF')
four_leads = ('I', 'II', 'III', 'V2')
three_leads = ('I', 'II', 'V2')
two_leads = ('I', 'II')
one_leads = ('I', )
lead_sets = (twelve_leads, six_leads, four_leads, three_leads, two_leads)
#lead_sets = (six_leads, four_leads, three_leads, two_leads, one_leads)
#lead_sets = (twelve_leads, six_leads, four_leads, three_leads, two_leads, one_leads)
#lead_sets = (two_leads, )
#lead_sets = (twelve_leads, )
#lead_sets = (one_leads,)

remote_flag = True

################################################################################
#
# Training model function
#
################################################################################

# Train your model. This function is *required*. Do *not* change the arguments of this function.
def training_code_image(data_directory, model_directory):
    # Find header and recording files.
    print('Finding header and recording files...')

    header_files, recording_files = find_challenge_files(data_directory)
    num_recordings = len(recording_files)

    if not num_recordings:
        raise Exception('No data was provided.')

    # Create a folder for the model if it does not already exist.
    if not os.path.isdir(model_directory):
        os.mkdir(model_directory)

    # Extract the classes from the dataset.
    print('Extracting classes...')

    #classes = set()
    #for header_file in header_files:
    #    header = load_header(header_file)
    #    classes |= set(get_labels(header))
    #if all(is_integer(x) for x in classes):
    #    classes = sorted(classes, key=lambda x: int(x)) # Sort classes numerically if numbers.
    #else:
    #    classes = sorted(classes) # Sort classes alphanumerically otherwise.
    #num_classes = len(classes)

    mapping_scoring = pd.read_csv('dx_mapping_scored.csv')
    classes = list(mapping_scoring['SNOMED CT Code'].values.astype(str))
    #classes = list(['270492004', '164889003', '164890007'])
    if all(is_integer(x) for x in classes):
        classes = sorted(classes, key=lambda x: int(x)) # Sort classes numerically if numbers.
    else:
        classes = sorted(classes) # Sort classes alphanumerically otherwise.
    num_classes = len(classes)

    # Extract features and labels from dataset.
    print('Extracting features and labels...')

    #data = np.zeros((num_recordings, 14), dtype=np.float32) # 14 features: one feature for each lead, one feature for age, and one feature for sex
    #data = np.array([])
    #labels = np.zeros((num_recordings, num_classes), dtype=np.bool) # One-hot encoding of classes
    #labels = np.array([])

    # Generate features
    if 0: 
        normal_class = ['426783006']
        equivalent_classes = [['713427006', '59118001'], ['284470004', '63593006'], ['427172004', '17338001']]
        for i in range(num_recordings):              
            print('    {}/{}...'.format(i+1, num_recordings))

            # Load header and recording.
            header = load_header(header_files[i])
            recording = load_recording(recording_files[i])

            local_data = utils.get_features_as_image(header, recording, twelve_leads, (224,224), image_type='mel')

            local_label = np.zeros((num_classes))
            current_labels = get_labels(header)
            for label in current_labels:
                if label in classes:
                    j = classes.index(label)
                    local_label[j] = 1

            if i == 0:
                data = local_data
                labels = local_label
            else:        
                data = np.dstack((data, local_data))
                labels = np.vstack((labels, local_label))
                
            if i%1000 == 1:
                # Save data                
                np.save(model_directory + 'data.npy', np.transpose(data, (2, 0, 1)))
                np.save(model_directory + 'labels.npy', labels)
        # Save data
        data = np.transpose(data, (2, 0, 1))
        np.save(model_directory + 'data.npy', data)
        np.save(model_directory + 'labels.npy', labels)
    else:
        # Load data
        data = np.load(model_directory + 'data.npy', allow_pickle=True)
        labels = np.load(model_directory + 'labels.npy', allow_pickle=True)

    # Train a model for each lead set.
    for leads in lead_sets:
        print('Training model for {}-lead set: {}...'.format(len(leads), ', '.join(leads)))

        filename = os.path.join(model_directory, get_model_filename(leads))

        feature_mult = [twelve_leads.index(lead) for lead in leads]
        lead_feature_len = int((data.shape[1]-2) / 12)
        feature_indices = np.array([])
        for lead_number in feature_mult:
            idxs = np.arange(lead_number*lead_feature_len, (lead_number+1)*lead_feature_len, 1)
            feature_indices = np.append(feature_indices, idxs)
        feature_indices = np.append(feature_indices, data.shape[1]-2)
        feature_indices = np.append(feature_indices, data.shape[1]-1)
        feature_indices = feature_indices.astype(int)
        features = data[:, feature_indices]

        imputer = SimpleImputer().fit(features)
        features = imputer.transform(features)
        classifier, scores, sensitivity, specificity = utils.xgboost_prepare_balanced(features, labels, optimize=False, random_state=42)        
        save_model(model_directory, leads, classes, imputer, classifier)    
        #np.save(os.path.join(model_directory, filename + 'scores.npy'), np.array(scores))
        #np.save(os.path.join(model_directory, filename + 'sensitivity.npy'), np.array(sensitivity))
        #np.save(os.path.join(model_directory, filename + 'specificity.npy'), np.array(specificity))

################################################################################

# Train your model. This function is *required*. Do *not* change the arguments of this function.
def training_code(data_directory, model_directory):
    # Find header and recording files.
    print('Finding header and recording files...')

    header_files, recording_files = find_challenge_files(data_directory)
    num_recordings = len(recording_files)

    if not num_recordings:
        raise Exception('No data was provided.')

    # Create a folder for the model if it does not already exist.
    if not os.path.isdir(model_directory):
        os.mkdir(model_directory)

    # Extract the classes from the dataset.
    print('Extracting classes...')

    #classes = set()
    #for header_file in header_files:
    #    header = load_header(header_file)
    #    classes |= set(get_labels(header))
    #if all(is_integer(x) for x in classes):
    #    classes = sorted(classes, key=lambda x: int(x)) # Sort classes numerically if numbers.
    #else:
    #    classes = sorted(classes) # Sort classes alphanumerically otherwise.
    #num_classes = len(classes)

    mapping_scoring = pd.read_csv('dx_mapping_scored.csv')
    classes = list(mapping_scoring['SNOMED CT Code'].values.astype(str))
    #classes = list(['270492004', '164889003', '164890007'])
    if all(is_integer(x) for x in classes):
        classes = sorted(classes, key=lambda x: int(x)) # Sort classes numerically if numbers.
    else:
        classes = sorted(classes) # Sort classes alphanumerically otherwise.
    num_classes = len(classes)

    # Extract features and labels from dataset.
    print('Extracting features and labels...')

    #data = np.zeros((num_recordings, 14), dtype=np.float32) # 14 features: one feature for each lead, one feature for age, and one feature for sex
    #data = np.array([])
    #labels = np.zeros((num_recordings, num_classes), dtype=np.bool) # One-hot encoding of classes
    #labels = np.array([])

    # Generate features
    if 1: 
        normal_class = ['426783006']
        equivalent_classes = [['713427006', '59118001'], ['284470004', '63593006'], ['427172004', '17338001']]

        fnames = pd.DataFrame()
        fnames['header_files'] = header_files
        fnames['recording_files'] = recording_files
        fnames['is_train'] = True        
        fnames['leads_amount'] = len(twelve_leads)

        #pool = joblib.Parallel(-1)
        #mapper = joblib.delayed(utils.load_row)
        #tasks = [mapper(row) for row in fnames.itertuples(False)]
        #res = pool(tasks)

        res = [utils.load_row(row) for row in fnames.itertuples(False)]
      
        data = np.array([row[0] for row in res])
        labels = np.array([row[1] for row in res])

        # Save data
        if not remote_flag:
            np.save(str(model_directory + 'data.npy'), data, allow_pickle=True)
            np.save(str(model_directory + 'labels.npy'), labels, allow_pickle=True)
    else:
        # Load data
        data = np.load(str(model_directory + 'data.npy'), allow_pickle=True)
        labels = np.load(str(model_directory + 'labels.npy'), allow_pickle=True)

    # Train a model for each lead set.
    for leads in lead_sets:
        print('Training model for {}-lead set: {}...'.format(len(leads), ', '.join(leads)))

        filename = os.path.join(model_directory, get_model_filename(leads))

        feature_mult = [twelve_leads.index(lead) for lead in leads]
        lead_feature_len = int((data.shape[1]-2) / 12)
        feature_indices = np.array([])
        for lead_number in feature_mult:
            idxs = np.arange(lead_number*lead_feature_len, (lead_number+1)*lead_feature_len, 1)
            feature_indices = np.append(feature_indices, idxs)
        feature_indices = np.append(feature_indices, data.shape[1]-2)
        feature_indices = np.append(feature_indices, data.shape[1]-1)
        feature_indices = feature_indices.astype(int)
        features = data[:, feature_indices]

        imputer = SimpleImputer().fit(features)
        features = imputer.transform(features)
        classifier, scores, sensitivity, specificity = utils.xgboost_prepare_balanced(features, labels, optimize=False, random_state=42)
        #classifier, scores, sensitivity, specificity = utils.rf_prepare_balanced(features, labels, optimize=False, random_state=42)
        save_model(model_directory, leads, classes, imputer, classifier)
        #np.save(os.path.join(model_directory, filename + 'scores.npy'), np.array(scores))
        #np.save(os.path.join(model_directory, filename + 'sensitivity.npy'), np.array(sensitivity))
        #np.save(os.path.join(model_directory, filename + 'specificity.npy'), np.array(specificity))

################################################################################
#
# Running trained model function
#
################################################################################

# Run your trained model. This function is *required*. Do *not* change the arguments of this function.
def run_model(model, header, recording):
    classes = model['classes']
    leads = model['leads']
    imputer = model['imputer']
    classifier = model['classifier']

    # Load features.    
    #data = utils.get_features_neurokit(header, recording, leads)
    data = utils.get_features_neurokit_mfcc2(header, recording, leads)
    #data = utils.get_features_neurokit_mfcc2(header, recording, leads)
    

    # Impute missing data.
    features = data.reshape(1, -1)
    features = imputer.transform(features)

    # Predict labels and probabilities.
    labels = np.array([])
    for mod in classifier:
        label = classifier[mod].predict(features)[0]
        labels = np.append(labels, label)    
    labels = np.asarray(labels, dtype=np.int)

    probabilities = np.array([])
    for mod in classifier:
        probability = classifier[mod].predict_proba(features)[0, 1]
        probabilities = np.append(probabilities, probability)    
    probabilities = np.asarray(probabilities, dtype=np.float32)

    return classes, labels, probabilities

################################################################################
#
# File I/O functions
#
################################################################################

# Save a trained model. This function is not required. You can change or remove it.
def save_model(model_directory, leads, classes, imputer, classifier):
    d = {'leads': leads, 'classes': classes, 'imputer': imputer, 'classifier': classifier}
    filename = os.path.join(model_directory, get_model_filename(leads))
    joblib.dump(d, filename, protocol=0)

# Load a trained model. This function is *required*. Do *not* change the arguments of this function.
def load_model(model_directory, leads):
    filename = os.path.join(model_directory, get_model_filename(leads))
    return joblib.load(filename)

# Define the filename(s) for the trained models. This function is not required. You can change or remove it.
def get_model_filename(leads):
    return 'model_' + '-'.join(sort_leads(leads)) + '.sav'

################################################################################
#
# Feature extraction function
#
################################################################################

def test_model_reduced_set(model_directory, data_directory, output_directory, leads_names=twelve_leads):
    # Find header and recording files.
    print('Finding header and recording files...')

    header_files, recording_files = find_challenge_files(data_directory)
    num_recordings = len(recording_files)

    if not num_recordings:
        raise Exception('No data was provided.')

    # Create a folder for the outputs if it does not already exist.
    if not os.path.isdir(output_directory):
        os.mkdir(output_directory)

    # Identify the required lead sets.
    required_lead_sets = set()
    for i in range(num_recordings):
        header = load_header(header_files[i])
        leads = get_leads(header)
        required_lead_sets.add(sort_leads(leads))

    required_lead_sets = leads_names
    # Load models.
    leads_to_model = dict()
    print('Loading models...')
    for leads in required_lead_sets:
        model = load_model(model_directory, leads) ### Implement this function!
        leads_to_model[leads] = model

    # Run model for each recording.
    print('Running model...')

    for i in range(num_recordings):
        print('    {}/{}...'.format(i+1, num_recordings))

        # Load header and recording.
        header = load_header(header_files[i])
        recording = load_recording(recording_files[i])
        #leads = get_leads(header)

        # Apply model to recording.
        model = leads_to_model[leads]
        classes, labels, probabilities = run_model(model, header, recording) ### Implement this function!

        # Save model outputs.
        recording_id = get_recording_id(header)
        head, tail = os.path.split(header_files[i])
        root, extension = os.path.splitext(tail)
        output_file = os.path.join(output_directory, root + '.csv')
        save_outputs(output_file, recording_id, classes, labels, probabilities)

    print('Done.')

# Test local model.
def test_model_reduced_set_(model_directory, data_directory, output_directory, amount_of_lead=12):
    # Load model.
    print('Loading models...')

    if amount_of_lead == 12:
        lead_model = load_model(model_directory, twelve_leads)
    elif amount_of_lead == 6:
        lead_model = load_model(model_directory, six_leads)
    elif amount_of_lead == 4:
        lead_model = load_model(model_directory, four_leads)        
    elif amount_of_lead == 3:
        lead_model = load_model(model_directory, three_leads)
    else:
        lead_model = load_model(model_directory, two_leads)

    # Find header and recording files.
    print('Finding header and recording files...')

    header_files, recording_files = find_challenge_files(data_directory)
    num_recordings = len(recording_files)

    if not num_recordings:
        raise Exception('No data was provided.')

    # Create a folder for the outputs if it does not already exist.
    if not os.path.isdir(output_directory):
        os.mkdir(output_directory)

    # Run model for each recording.
    print('Running model...')

    for i in range(num_recordings):
        print('    {}/{}...'.format(i+1, num_recordings))

        # Load header and recording.
        header = load_header(header_files[i])
        recording = load_recording(recording_files[i])
        leads = get_leads(header)

        # Apply model to recording.
        if all(lead in leads for lead in twelve_leads) and amount_of_lead == 12:
            classes, labels, probabilities = run_twelve_lead_model(lead_model, header, recording)
        elif all(lead in leads for lead in six_leads) and amount_of_lead == 6:
            classes, labels, probabilities = run_six_lead_model(lead_model, header, recording)
        elif all(lead in leads for lead in three_leads) and amount_of_lead == 3:
            classes, labels, probabilities = run_three_lead_model(lead_model, header, recording)
        elif all(lead in leads for lead in two_leads)  and amount_of_lead == 2:
            classes, labels, probabilities = run_two_lead_model(lead_model, header, recording)
        else:
            raise NotImplementedError('No model is implemented for the lead set {}.'.format(', '.join(leads)))

        # Save model outputs.
        head, tail = os.path.split(header_files[i])
        root, extension = os.path.splitext(tail)
        output_file = os.path.join(output_directory, root + '.csv')
        save_outputs(output_file, classes, labels, probabilities)

    print('Done.')

# Extract features from the header and recording. This function is not required. You can change or remove it.
def get_features_(header, recording, leads):
    # Extract age.
    age = get_age(header)
    if age is None:
        age = float('nan')

    # Extract sex. Encode as 0 for female, 1 for male, and NaN for other.
    sex = get_sex(header)
    if sex in ('Female', 'female', 'F', 'f'):
        sex = 0
    elif sex in ('Male', 'male', 'M', 'm'):
        sex = 1
    else:
        sex = float('nan')

    # Reorder/reselect leads in recordings.
    recording = choose_leads(recording, header, leads)

    # Pre-process recordings.
    adc_gains = get_adc_gains(header, leads)
    baselines = get_baselines(header, leads)
    num_leads = len(leads)
    for i in range(num_leads):
        recording[i, :] = (recording[i, :] - baselines[i]) / adc_gains[i]

    # Compute the root mean square of each ECG lead signal.
    rms = np.zeros(num_leads)
    for i in range(num_leads):
        x = recording[i, :]
        rms[i] = np.sqrt(np.sum(x**2) / np.size(x))

    return age, sex, rms
